//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
//
//     fourier.pas  -  Don Cross <dcross <at> intersrv <dot> com>
//
//     This is a Turbo Pascal Unit for calculating the Fast Fourier Transform
//     (FFT) and the Inverse Fast Fourier Transform (IFFT).
//     Visit the following URL for the latest version of this code.
//     This page also has a C/C++ version, and a brief discussion of the
//     theory behind the FFT algorithm.
//
//        http://www.intersrv.com/~dcross/fft.html#pascal [!!Dead link!!]
//
//     Revision history [most recent first]:
//
// 2008 May 12 [Nick Currie]
//     All code transcoded to IA-32/x87. FFT_Real implemented to allow calculation
//     of a packed Real array of length 2N to be calculated in a Complex array of
//     length N, with subsequent untangling. ReversedBits, Reversed_Bits, Ai & Ar
//     lookup tables implemented. Calc_Frequency, FFT_Integer and
//     FFT_Integer_Cleanup removed as not used in lossyWAV.
//
// 1996 December 11 [Don Cross]
//     Improved documentation of the procedure CalcFrequency.
//     Fixed some messed up comments in procedure IFFT.
//
// 1996 December 6 [Don Cross]
//     Made procedure 'fft_integer' more efficient when buffer size changes
//     in successive calls:  the buffer is now only resized when the input
//     has more samples, not a differing number of samples.
//     Also changed the way 'fft_integer_cleanup' works so that it is
//     more "bullet-proof".
//
// 1996 December 4 [Don Cross]
//     Adding the procedure 'CalcFrequency', which calculates the FFT
//     at a specific frequency index p=0..n-1, instead of the whole
//     FFT.  This is O(n^2) instead of O(n*log(n)).
//
// 1996 November 30 [Don Cross]
//     Adding a routine to allow FFT of an input array of integers.
//     It is called 'fft_integer'.
//
// 1996 November 18 [Don Cross]
//     Added some comments.
//
// 1996 November 17 [Don Cross]
//     Wrote and debugged first version.
//
//==============================================================================

//==============================================================================
unit nfft;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8}
//==============================================================================
interface
//==============================================================================
uses
  ntypes;

Const
  MaxFFTLength    = 1024;
  MaxFFTBitLength = 10;

Type
  tFFT_Array = Packed Record
    Case Integer of
      0 : (Complex : Array[0..MaxFFTLength-1] of TComplex);
      1 : (Real    : Array[0..(MaxFFTLength-1)*2] of Double);
  End;

Var
  nfft_data_aligner_1 : array[0..0] of integer; //data alignment (optimised by iteration)
  FFT_Array           : tFFT_Array;
  nfft_data_aligner_2 : array[0..0] of integer; //data alignment (optimised by iteration)

//------------------------------------------------------------------------------
// Set the number of samples for FFT analysis;
//------------------------------------------------------------------------------
function SetFFTLength(SFL_Samples : Integer) : Integer;

//------------------------------------------------------------------------------
//  Calculates the (in place) Fast Fourier Transform of the array of complex
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array.
//------------------------------------------------------------------------------
procedure FFT_Complex;

//------------------------------------------------------------------------------
//  Calculates the (in place) Fast Fourier Transform of the packed array of real
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array, then post-processed the results to unpack the
//  Real output.
//------------------------------------------------------------------------------
procedure FFT_Real;

//------------------------------------------------------------------------------
//  Calculates the (in place) Inverse Fast Fourier Transform of the array of
//  complex numbers represented by FFT_Input to produce the output
//  complex numbers in FFT_Array.
//------------------------------------------------------------------------------
procedure IFFT_Complex;

//------------------------------------------------------------------------------

//==============================================================================
implementation
//==============================================================================

var
  RB_Arr         : TwoPowerDoubleRecord;
  AP_Arr, AN_Arr : TwoPowerComplexRecord;

  ReversedBits   : array[1..MaxFFTBitLength] of PtrIntegerVector   = (@RB_Arr.p1,@RB_Arr.p2,@RB_Arr.p3,@RB_Arr.p4,@RB_Arr.p5,@RB_Arr.p6,@RB_Arr.p7,@RB_Arr.p8,@RB_Arr.p9,@RB_Arr.pA);//,@RB_Arr.pB,@RB_Arr.pC,@RB_Arr.pD,@RB_Arr.pE,@RB_Arr.pF);
  AP_Ptr         : Array[1..MaxFFTBitLength+1] of PtrComplexVector = (@AP_Arr.p0,@AP_Arr.p1,@AP_Arr.p2,@AP_Arr.p3,@AP_Arr.p4,@AP_Arr.p5,@AP_Arr.p6,@AP_Arr.p7,@AP_Arr.p8,@AP_Arr.p9,@AP_Arr.pA);//,@AP_Arr.pB,@AP_Arr.pC,@AP_Arr.pD,@AP_Arr.pE,@AP_Arr.pF);
  AN_Ptr         : Array[1..MaxFFTBitLength+1] of PtrComplexVector = (@AN_Arr.p0,@AN_Arr.p1,@AN_Arr.p2,@AN_Arr.p3,@AN_Arr.p4,@AN_Arr.p5,@AN_Arr.p6,@AN_Arr.p7,@AN_Arr.p8,@AN_Arr.p9,@AN_Arr.pA);//,@AN_Arr.pB,@AN_Arr.pC,@AN_Arr.pD,@AN_Arr.pE,@AN_Arr.pF);
  A_Ptr          : PtrIntegerVector = @AP_Ptr;

  Alpha                 : Double;
  Beta                  : Double;
  NumSamples_Reciprocal : Double;

  NumSamples            : Integer;
  BlockSize             : Integer;
  BlockBitLen           : Integer;
  NumberOfBitsNeeded    : Integer;

//  nfft_dummy            : array[0..3] of integer;

  function SetFFTLength(SFL_Samples : Integer) : integer; assembler;
  asm
                //store NumComplexSamples
                mov   eax,SFL_Samples

                //find highest set bit
                bsr   edx,eax
                jz    @@01
                //find lowest set bit
                bsf   ecx,eax
                jz    @@01
                //if highest=lowest then poweroftwo
                cmp   edx,ecx
                je    @@02
                //else
      @@01:     xor   eax,eax
                xor   edx,edx

      @@02:     mov   NumberOfBitsNeeded,edx
                mov   NumSamples,eax
  end;

  procedure Shuffle_In_Place; assembler;
  asm
                push   ebx
                push   edi

                mov    edx,numberofbitsneeded
                mov    edx,dword ptr reversedbits[edx*pointer_size-pointer_size]
                mov    ecx,NumSamples
                xor    edi,edi
                xor    ebx,ebx
                jmp    @@S2

      @@S1:       mov    eax,dword ptr [edx+edi*integer_size]
                  cmp    eax,ebx
                  jle    @@S2

                  fld    qword ptr FFT_Array[ebx]
                  fld    qword ptr FFT_Array[ebx+real_size]

                  fld    qword ptr FFT_Array[eax]
                  fld    qword ptr FFT_Array[eax+real_size]

                  fstp   qword ptr FFT_Array[ebx+real_size]
                  fstp   qword ptr FFT_Array[ebx]

                  fstp   qword ptr FFT_Array[eax+real_size]
                  fstp   qword ptr FFT_Array[eax]

      @@S2:       add    edi,1
                  add    ebx,complex_size
                  sub    ecx,1
                  jg     @@S1

                pop    edi
                pop    ebx
  end;

  procedure FFT_Complex; assembler;
  asm
                push   ebx
                push   esi
                push   edi
                push   ebp

                call   Shuffle_In_Place

                mov    ebx,1                                            // blockend
                mov    blockbitlen,1                                    // blockbitlen
                mov    blocksize,2                                      // blocksize

      @@FC3:      xor    ebp,ebp                                        //i

      @@FC4:        mov    edi,ebp                                      //j
                    xor    esi,esi                                      //n

      @@FC5:          mov     eax,blockbitlen

                      mov     edx,dword ptr A_Ptr
                      mov     edx,dword ptr [edx+eax*pointer_size-pointer_size]
                      mov     eax,esi
                      shl     eax,complex_shift
                      fld     qword ptr [edx+eax+real_size]             //a.im:=a_ptr[blockbitlen]^[n].Im
                      fld     qword ptr [edx+eax]                       //a.re:=a_ptr[blockbitlen]^[n].Re

                      mov     ecx,edi                                   //j
                      mov     edx,edi
                      add     edx,ebx                                   //k:=j+blockend
                      shl     ecx,complex_shift
                      shl     edx,complex_shift

                      // load FFT_Array[k] into FPU
                      fld     qword ptr FFT_Array[edx+real_size]       // FFT_Array[k].im
                      fld     qword ptr FFT_Array[edx]                 // FFT_Array[k].re

                      // T.re := A.re * FFT_Array.Complex[K].Re - A.im * FFT_Array.Complex[K].Im;
                      fld     st(0)
                      fmul    st,st(3)
                      fld     st(2)
                      fmul    st,st(5)
                      fsub                                              // t.re
                      fxch    st(4)

                      // T.im := A.re * FFT_Array.Complex[K].Im + A.im * FFT_Array.Complex[K].Re;
                      fmul
                      fxch    st(2)
                      fmul
                      fadd                                              // t.im

                      fld     qword ptr FFT_Array[ecx+real_size]       // FFT_Array[j].im
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx+real_size]       // FFT_Array[k].im:=FFT_Array[j].im-t.im
                      fadd
                      fstp    qword ptr FFT_Array[ecx+real_size]       // FFT_Array[j].im:=FFT_Array[j].im+t.im

                      fld     qword ptr FFT_Array[ecx]                 // FFT_Array[j].re
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx]                 // FFT_Array[k].re:=FFT_Array[j.re]-t.re
                      fadd
                      fstp    qword ptr FFT_Array[ecx]                 // FFT_Array[j].re:=FFT_Array[j.re]+t.re

                      add    edi,1                                      // j:=j+1
                      add    esi,1                                      // n:=n+1
                      cmp    esi,ebx
                      jl     @@FC5

                    mov    eax,blocksize
                    add    eax,ebp
                    mov    ebp,eax                                      // fft_i:=fft_i+blocksize
                    cmp    eax,NumSamples
                    jl     @@FC4

                  mov    eax,blocksize
                  mov    ebx,eax                                        // blockend:=blockend+blocksize;
                  add    blocksize,eax                                  // blocksize:=2*blocksize;

                  add    blockbitlen,1
                  mov    eax,blockbitlen
                  cmp    eax,numberofbitsneeded
                  jle    @@FC3

                pop    ebp
                pop    edi
                pop    esi
                pop    ebx
  end;

  procedure FFT_Real; Assembler;
  Asm
                push    ebx

                call    FFT_Complex

                mov     edx,numberofbitsneeded
                mov     edx,dword ptr ap_ptr[edx*pointer_size]

                Mov     ebx,NumSamples
                shl     ebx,complex_shift

                Mov     ecx,ebx
                shr     ecx,1

      @@FR1:    mov     eax,ebx
                sub     eax,ecx
                                                                         // -
                fld     qword ptr powersoftwoin[1*real_size]             //#0#
                fld     qword ptr FFT_Array[ecx]                         //#1#
                fld     qword ptr FFT_Array[ecx+real_size]               //#2#
                fld     qword ptr FFT_Array[eax]                         //#3#
                fld     qword ptr FFT_Array[eax+real_size]               //#4#

                fld     st(2)                                 //temp1.im    5
                fadd    st,st(1)                              //temp2.im    5
                fmul    qword ptr [edx+ecx]                   //ar          5
                fld     st(4)                                 //temp1.re    6
                fsub    st,st(3)                              //temp2.re    6
                fmul    qword ptr [edx+ecx+real_size]         //ai          6
                fadd                                          //T.re       #5#

                fld     st(4)                                 //temp1.re    6
                fsub    st,st(3)                              //temp2.re    6
                fmul    qword ptr [edx+ecx]                   //ar          6
                fld     st(4)                                 //temp1.im    7
                fadd    st,st(3)                              //temp2.im    7
                fmul    qword ptr [edx+ecx+real_size]         //ai          7
                fsub                                          //T.im       #6#

                fld     st(5)                                 //temp1.re    7
                fadd    st,st(4)                              //temp2.re    7
                fadd    st,st(2)                              //T.re        7
                fmul    st,st(7)                              //0.5         7
                fstp    qword ptr FFT_Array[ecx]                 //6

                fld     st(4)                                 //temp1.im    7
                fsub    st,st(3)                              //temp2.im    7
                fsub    st,st(1)                              //T.Im        7
                fmul    st,st(7)                              //0.5         7
                fstp    qword ptr FFT_Array[ecx+real_size]       //6

                fld     st(2)                                 //temp2.im    7
                fsub    st,st(5)                              //temp1.im    7
                fsubr                                         //T.Im        6
                fmul    st,st(6)                              //0.5         6
                fstp    qword ptr FFT_Array[eax+real_size]               //#5#

                fld     st(4)                                 //temp1.re    6
                fadd    st,st(3)                              //temp2.re    6
                fsubr                                         //T.re        5
                fmul    st,st(5)                              //0.5         5
                fstp    qword ptr FFT_Array[eax]                         //#4#

                fstp    st                                               //#3#
                fstp    st                                               //#2#
                fstp    st                                               //#1#
                fstp    st                                               //#0#
                fstp    st                                               //#-#

                sub     ecx,complex_size
                jg      @@FR1

                fld     qword ptr FFT_Array[0*real_size]
                fld     qword ptr FFT_Array[1*real_size]
                fadd
                fstp    qword ptr FFT_Array[0*real_size]
                fldz
                fstp    qword ptr FFT_Array[1*real_size]

                pop     ebx
  end;

  procedure IFFT_Complex; assembler;
    asm
                mov    A_Ptr,offset AN_Ptr

                call   FFT_Complex

                mov    A_Ptr,offset AP_Ptr

                fld    NumSamples_reciprocal
                mov    ecx,NumSamples
                shl    ecx,complex_shift

      @@01:     sub    ecx,complex_size
                jle    @@02

                fld    qword ptr FFT_Array[ecx]
                fmul   st,st(1)
                fstp   qword ptr FFT_Array[ecx]

                fld    qword ptr FFT_Array[ecx+real_size]
                fmul   st,st(1)
                fstp   qword ptr FFT_Array[ecx+real_size]

                jmp    @@01

      @@02:     fstp   st
    end;

//==============================================================================
begin
//==============================================================================
    asm
                push  ebx
                push  ebp
                push  edi
                push  esi

                xor   eax,eax
                mov   dword ptr RB_Arr,eax
                mov   dword ptr AP_Arr,eax
                mov   dword ptr AN_Arr,eax

                mov   dword ptr nfft_data_aligner_1,eax
                mov   dword ptr nfft_data_aligner_2,eax

      // Zero fft_input & FFT_Array arrays
                mov   ecx,maxfftlength
                add   ecx,ecx
                fldz

      @@00:     //fst   qword ptr FFT_Input[ecx*real_size-real_size]
                fst   qword ptr FFT_Array[ecx*real_size-real_size]
                sub   ecx,1
                jnz   @@00

                fstp  st

      // Fill ReversedBits lookup tables (shifted to take into account complex_size) for all FFT lengths
                mov   edi,MaxFFTBitLength
      @@01:     mov   esi,dword ptr PowersOfTwoM1[edi*int64_size]
                mov   ebx,dword ptr ReversedBits[edi*pointer_size-pointer_size]

      @@02:     mov   edx,edi
                mov   ecx,esi
                xor   eax,eax

      @@03:       rcr   ecx,1
                  rcl   eax,1

                  sub   edx,1
                  jg    @@03

                shl   eax,complex_shift
                mov   dword ptr [ebx+esi*integer_size],eax

                sub   esi,1
                jge   @@02

                sub   edi,1
                jg    @@01

                mov    ebx,1

      @@04:     mov    eax,ebx
                fld    qword ptr powersoftwoin[eax*real_size]
                fldpi
                fmul
                // Delta_Angle:=Pi*PowersOfTwoIn[j];

                fld    st
                fsin
                fmul   st,st
                fadd   st,st
                // Alpha:=2*Sqr(Sin(Delta));
                fstp   alpha

                fadd   st,st
                fsin
                // Beta:=Sin(2*Delta);
                fstp   beta

      // Initialise Ais[0] & Ars[0] values.

                mov    edx,ebx
                sub    edx,1
                shl    edx,integer_shift

                mov    ecx,dword Ptr AP_Ptr[edx]
                fld1
                fstp   qword ptr [ecx]           //AP_Arr[blockbitlen]^[0].Re:=1;
                fldz
                fstp   qword ptr [ecx+real_size] //AP_Arr[blockbitlen]^[0].Im:=0;

                mov    ecx,dword Ptr AN_Ptr[edx]
                fld1
                fstp   qword ptr [ecx]           //AN_Arr[blockbitlen]^[0].Re:=1;
                fldz
                fstp   qword ptr [ecx+real_size] //AN_Arr[blockbitlen]^[0].Im:=0;

      // Calculate Ai and Ar lookup table.
                cmp    ebx,2
                jl     @@06

                  mov    eax,ebx
                  mov    ecx,dword ptr powersoftwo[eax*int64_size-int64_size]

                  mov    edx,1
                  mov    edi,complex_size

                  fld1
                  fsub   alpha
                  fld    beta

      @@05:         mov    esi,dword ptr AP_Ptr[eax*pointer_size-pointer_size]
                    mov    ebp,dword ptr AN_Ptr[eax*pointer_size-pointer_size]

                    fld    qword ptr [esi+edi-complex_size]            //re
                    fmul   st,st(2)
                    fld    qword ptr [esi+edi-complex_size+real_size]  //im
                    fmul   st,st(2)
                    fadd
                    fstp   qword ptr [esi+edi]                         //re

                    fld    qword ptr [esi+edi-complex_size+real_size]  //im
                    fmul   st,st(2)
                    fld    qword ptr [esi+edi-complex_size]            //re
                    fmul   st,st(2)
                    fsub
                    fstp   qword ptr [esi+edi+real_size]               //im

                    fld    qword ptr [ebp+edi-complex_size]            //re
                    fmul   st,st(2)
                    fld    qword ptr [ebp+edi-complex_size+real_size]  //im
                    fmul   st,st(2)
                    fsub
                    fstp   qword ptr [ebp+edi]                         //re

                    fld    qword ptr [ebp+edi-complex_size+real_size]  //im
                    fmul   st,st(2)
                    fld    qword ptr [ebp+edi-complex_size]            //re
                    fmul   st,st(2)
                    fadd
                    fstp   qword ptr [ebp+edi+real_size]               //im

                    add    edi,complex_size
                    add    edx,1
                    cmp    edx,ecx
                    jl     @@05

                  fstp   st
                  fstp   st

      @@06:     //end of the loop
                cmp    ebx,maxfftbitlength
                jg     @@07

                add    ebx,1
                jmp    @@04

      @@07:     pop    esi
                pop    edi
                pop    ebp
                pop    ebx
    End;
//==============================================================================
// Unit nfft end;
//==============================================================================
end.
//==============================================================================

