//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nparameter;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8}
//==============================================================================
interface
//==============================================================================

Const
  version_string='lossyWAV 1.0.0b';

var
  Parameter_String : string[255];

  procedure lossyWAVerror  (lwe_string : string; lwe_value : integer);
  procedure lossyWAVwarning(lww_string : string);

  function  DirInp                : string;   // name of the source file folder.
  function  DirOut                : String;   // name of the output file folder.

  function  wavInp                : string;   // name of the input wav file & directory
  function  wavOut                : string;   // name of the input wav file & directory

  function  lossyInp              : String;   // name of the lossy input wav file.
  function  lossyOut              : string;   // name of the lossy output wav file based on the input file name.

  function  lwcdfInp              : String;   // name of the lwcdf input wav file.
  function  lwcdfOut              : string;   // name of the lwcdf output wav file based on the input file name.

  function  WAVFileName           : string;   // input filename only;
  function  lossyFileName         : string;   // output filename only;
  function  lwcdfFileName         : string;   // correction file output filename only;

  function  quality               : Double;   // quality level. default=2;
  function  forcing               : Integer;  // forcibly over-write output file if it exists.
  function  correct               : Integer;  // write correction file if correct=1.
  function  checking              : Integer;

  function  noclips               : Integer;  // set allowable clips to zero for all quality presets.
  function  merging               : Integer;
  function  dither                : Integer;  // enable TPDF dither when noise shaping.
  function  silent                : Integer;  // enable silent running.

  function  verbosity             : Integer;  // verbosity level. default=false;
  function  warnings              : Integer;  // warnings switch state. default=false;
  function  priority              : Integer;  // below or low_priority process priority switch;
  function  analyses              : Integer;  // user selectable extra fft analysis.

  function  spf_string            : String;   // return manually input spreading functions.
  function  fft_string            : String;   // return manually input FFT selection;
  function  minbits               : Double;
  function  snr                   : Double;   // signal to noise ratio.
  function  noisethreshold        : Double;   // noise_threshold_shift value.

  function  shaping               : Double;   // enable optional fixed noise shaping; default=off.
  function  autoshape             : Integer;  // enable optional auto-noise-shaping factor.

  function  fft32                 : Integer;  // enable 32 sample FFT analysis for improved impulse detection.
  function  lowpass               : Double;   // modify upper bound of frequency range analysed.
  function  skewing               : Double;   // modify skewing amplitude.
  function  scaling               : Double;   // scaling factor from WaveGain;
  function  detailledoutput       : Integer;  // debug switch state. default=false

//==============================================================================
implementation
//==============================================================================

uses
  SysUtils,
  Math,
  nwav,
  ntypes;

//==============================================================================
//nparameter constants, variables & types
//==============================================================================
const
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAVHead1=version_string+', Copyright (C) 2007,2008 Nick Currie. Copyleft.' + crlf;
  lossyWAVhead2= 'This is free software under the GNU GPLv3+ license; There is NO WARRANTY, to'+crlf
                +'the extent permitted by law. <http://www.gnu.org/licenses/> for details.'+crlf;
  lossyWAVheader=lossyWAVhead1+lossyWAVhead2;
  lossyWAVhead3= 'This program is free software: you can redistribute it and/or modify it under'+crlf+
                 'the terms of the GNU General Public License as published by the Free Software'+crlf+
                 'Foundation, either version 3 of the License, or (at your option) any later'+crlf+
                 'version.'+crlf+crlf+
                 'This program is distributed in the hope that it will be useful,but WITHOUT ANY'+crlf+
                 'WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A'+crlf+
                 'PARTICULAR PURPOSE.  See the GNU General Public License for more details.'+crlf+crlf+
                 'You should have received a copy of the GNU General Public License along with'+crlf+
                 'this program.  If not, see <http://www.gnu.org/licenses/>.'+crlf;
var
    parmError                 : string;
    WavInpDir                 : string;
    WavOutDir                 : string;

    wavName                   : String;
    lossyName                 : String;
    lwcdfName                 : String;

    spf_int                   : String;
    fft_set_int               : String;

    noisethreshold_int        : Double;
    snr_int                   : Double;
    shaping_int               : Double;
    skewing_int               : Double;

    scaling_int               : Double;
    lowpass_int               : Double;
    quality_int               : Double;
    minbits_int               : Double;

    noclips_int               : Integer;
    detail_int                : Integer;
    analyses_int              : Integer;
    fft32_int                 : Integer;

    verbosity_int             : Integer;
    warnings_int              : Integer;
    forcing_int               : Integer;
    NextParamStrParamNr       : Integer; // internal param counter for nextParamStr

    priority_int              : Integer;
    correct_int               : Integer;
    checking_int              : Integer;
    merging_int               : Integer;

    autoshape_int             : Integer;
    help_int                  : Integer;
    dither_int                : Integer;
    silent_int                : Integer;

    nparameter_dummy_integer  : array[0..3] of integer;

  procedure lossyWAVerror(lwe_string : string; lwe_value : integer);
  begin
    if (silent=0) then
      write('%lossyWAV Error% : ',lwe_string,crlf);
    halt(lwe_value);
  end;

  procedure lossyWAVwarning(lww_string : string);
  begin
    if (warnings=1) and (silent=0) then
      write('%lossyWAV Warning% : ',lww_string,crlf);
  end;

  procedure parmsError;
  begin
    if (paramcount>0) and (help_int=0) then
      write(parmerror,crlf)
    else
    begin
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(     'Usage   : lossyWAV <input wav file> <options>'+
            crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'Example : lossyWAV musicfile.wav');
      write(crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'Quality Options:');
      write(crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'-q, --quality <n>   quality preset (10=highest quality, 0=lowest bitrate;'+
            crlf+'                    -q 5 is generally accepted to be transparent)'+
            crlf+'                    default=-q 5.');
      write(crlf+
            crlf+'Standard Options:'+
            crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'-c, --check         check if WAV file has already been processed; default=off.'+
            crlf+'                    errorlevel=16 if already processed, 0 if not.');
      write(crlf+'-C, --correction    write correction file for processed WAV file; default=off.');
      write(crlf+'-f, --force         forcibly over-write output file if it exists; default=off.');
      write(crlf+'-h, --help          display help.');
      write(crlf+'-L, --longhelp      display extended help.');
      write(crlf+'-M, --merge         merge existing lossy.wav and lwcdf.wav files.');
      write(crlf+'-N, --noclips       set allowable number of clips / channel / codec block to 0;'+
            crlf+'                    default=3,3,3,3,2,1,0,0,0,0,0 (-q 0 to -q 10)');
      write(crlf+'-o, --outdir <dir>  destination directory for the output file(s).');
      write(crlf+'-v, --version       display the lossyWAV version number.');
      if help_int>1 then
      begin
      write(crlf+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'Advanced Options:'+
            crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'-a, --analyses <n>  select number of FFT analysis lengths to use; (2<=n<=5);'+
            crlf+'                    default=2, i.e. 64 sample and 1024 sample FFT analyses;'+
            crlf+'                    (3=+128 sample FFT; 4=+256 sample FFT; 5=+512 sample FFT).');
      write(crlf+'-F, --fft32         enable 32 sample FFT for improved impulse detection;'+
            crlf+'                    defaults: -q 0 to 2=off; -q 3 to 10=on.');
      write(crlf+'-m, --minbits <n>   select minimum bits to keep (0.00<=n<=8.00);'+
            crlf+'                    default=2.9,2.95,3,3.125,3.25,3.375,3.5,3.625,3.75,3.875,4.');
      write(crlf+'-s, --scale <n>     scaling factor from WaveGain, etc; default=1.000000; n<>0!');
      write(crlf+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'System Options:'+
            crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'-d, --detail        enable detailed output mode');
      write(crlf+'-n, --nowarnings    suppress lossyWAV warnings.');
      write(crlf+'-Q, --quiet         significantly reduce screen output.');
      write(crlf+'    --silent        no screen output.');
      write(crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'-b, --below         set process priority to below normal.');
      write(crlf+'-l, --low           set process priority to low.');
      end;
      write(crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
      write(crlf+'Special thanks:'+crlf);
      write(crlf+'David Robinson for the method itself and motivation to implement it.');
      write(crlf+'Don Cross for the Complex-FFT algorithm used.');
      write(crlf+'Horst Albrecht for valuable tuning input and feedback.');
//      write(crlf+'Sebastian Gesemann for helping me to implement noise shaping properly.');
      write(crlf);//345678901234567890123456789012345678901234567890123456789012345678901234567890
    end;
  end;

  procedure initNextParamStr;
  begin
   NextParamStrParamNr:=0;
  end;

  function nextParamStr(var parm: string): Integer;
  begin
    if NextParamStrParamNr>=paramCount then
    begin
      nextParamStr:=0;
      exit;
    end;
    NextParamStrParamNr:=NextParamStrParamNr + 1;
    parm:=ParamStr(NextParamStrParamNr);
    nextParamStr:=1;
  end;

  procedure parmerror_multiple_selection;
  begin
    lossyWAVerror('Multiple '+parmerror+' switches given.',$31);
  end;

  procedure parmerror_mutually_incompatible_selection;
  begin
    lossyWAVerror('Switches '+parmerror+' are incompatible.',$31);
  end;

  procedure parmerror_no_value_given;
  begin
    lossyWAVerror('No '+parmerror+' value given.',$31);
  end;

  procedure parmerror_val_error;
  begin
    lossyWAVerror('Error evaluating '+parmerror+' value given.',$31);
  end;

  procedure writeversionstring;
  begin
    write(version_string,crlf);
    halt(0);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Integer); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+inttostr(cpv_low)+'<=n<='+inttostr(cpv_high),$31);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Double); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+floattostrf(cpv_low,ffFixed,0,2)+'<=n<='+floattostrf(cpv_high,ffFixed,0,2),$31);
  end;

  function check_quality(s : string) : boolean;
  var
    valError : Integer;
  begin
    check_quality:=false;

    if (s='-v') or (s='--version') then
      writeversionstring;

    if (s='-h') or (s='--help') then
    begin
      write(lossyWAVhead1,crlf);
      write(lossyWAVhead3,crlf);
      help_int:=1;
      parmserror;
      halt(0);
    end;

    if (s='-L') or (s='--longhelp') then
    begin
      write(lossyWAVhead1,crlf);
      write(lossyWAVhead3,crlf);
      help_int:=2;
      parmserror;
      halt(0);
    end;

    if (s='-q') or (s='--quality') then
    begin
      parmerror:='quality preset';

      if quality_int<>-1 then
        parmerror_multiple_selection;

      if nextParamStr(s)=0 then
        parmerror_no_value_given;

      Val(s, quality_Int, valError);
      if valError <> 0 then
        parmerror_val_error;

      quality_int:=abs(round(quality_int*10000)/10000);

      check_permitted_values(quality_int,0,10);
      check_quality:=true;
    end;

    if (s='-a') or (s='--analyses') then
    begin
      parmerror:='analyses';

      if analyses_int<>0 then
        parmerror_multiple_selection;

      if nextParamStr(s)=0 then
        parmerror_no_value_given;

      Val(s,analyses_int,valError);
      if valerror<>0 then
        parmerror_val_error;

      check_permitted_values(analyses_int,2,5);

      check_quality:=true;
    end;

    if (s='-F') or (s='--fft32') then
    begin
      parmerror:='fft32';

      if fft32_int<>0 then
        parmerror_multiple_selection;

      fft32_int:=1;

      check_quality:=true;
    end;
  end;

  function check_system(s : string) : boolean;
  begin
    check_system:=false;

    if (s='-b') or (s='--below') then
    begin
      parmerror:='process priority';
      if (priority_int>0) then
        parmerror_multiple_selection;
      priority_int:=1;
      check_system:=true;
    end;

    if (s='-l') or (s='--low') then
    begin
      parmerror:='process priority';
      if (priority_int>0) then
        parmerror_multiple_selection;
      priority_int:=2;
      check_system:=true;
    end;
  end;

  function check_output(s : string) : boolean;
  begin
    check_output:=false;

    if (s='-d') or (s='--detail') then
    begin
      parmerror:='detail';
      if (detail_int=1) then
        parmerror_multiple_selection;
      detail_int:=1;
      check_output:=true;
    end;

    if (s='-Q') or (s='--quiet') then
    begin
      parmerror:='reduced output';
      if verbosity_int=0 then
        parmerror_multiple_selection;
      verbosity_int:=0;
      check_output:=true;
    end;

    if             (s='--silent') then
    begin
      parmerror:='silent running';
      if silent_int=1 then
        parmerror_multiple_selection;
      silent_int:=1;
      check_output:=true;
    end;

    if (s='-n') or (s='--nowarnings') then
    begin
      parmerror:='no warnings';
      if warnings_int=0 then
        parmerror_multiple_selection;
      warnings_int:=0;
      check_output:=true;
    end;

    if (s='-c') or (s='--check') then
    begin
      parmerror:='check';
      if checking_int=1 then
        parmerror_multiple_selection;
      checking_int:=1;
      check_output:=true;
    end;

    if (s='-M') or (s='--merge') then
    begin
      parmerror:='merge';
      if merging_int=1 then
        parmerror_multiple_selection;
      merging_int:=1;
      check_output:=true;
    end;
  end;

  function check_file_output(s : string) : boolean;
  begin
    check_file_output:=false;

    if (s='-f') or (s='--force') then
    begin
      parmerror:='forced over-write';
      if forcing_int=1 then
        parmerror_multiple_selection;
      forcing_int:=1;
      check_file_output:=true;
    end;

    if (s='-o') or (s='--outdir') then
    begin
      parmerror:='destination folder';
      if (WAVOutDir <> '') then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      WAVOutDir:=s;
      if not((WAVOutDir[length(WAVOutDir)]='\') or (WAVOutDir[length(WAVOutDir)]=':')) then
        WAVOutdir:=WAVOutdir+'\';
      check_file_output:=true;
    end;

    if (s='-C') or (s='--correction') then
    begin
      parmerror:='correction file';
      if correct_int=1 then
        parmerror_multiple_selection;
      correct_int:=1;
      check_file_output:=true;
    end;
  end;

  function check_advanced(s : string) : boolean;
  var
    valError : Integer;
  begin
    check_advanced:=false;

    if (s='-m') or (s='--minbits') then
    begin
      parmerror:='minimum bits to keep modifier';
      if minbits_int<>-1 then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, minbits_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      minbits_int:=round(minbits_int*100)/100;
      check_permitted_values(minbits_int,0.0,8.0);
      check_advanced:=true;
    end;

    if (s='-s') or (s='--scale') then
    begin
      parmerror:='scaling factor';
      if scaling_int<>-1 then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, scaling_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      scaling_int:=round(scaling_int*1000000)/1000000;
      check_permitted_values(scaling_int,0.0,1.0);
      check_advanced:=true;
    end;

    if (s='-N') or (s='--noclips') then
    begin
      parmerror:='no allowable clips';
      if noclips_int<>0 then
        parmerror_multiple_selection;
      noclips_int:=1;
      check_advanced:=true;
    end;
  end;

  function getParms: Integer;
  var
    s        : string;
    colonpos : Integer;
    slashpos : Integer;
  begin
    getParms:=0;
    spf_int:='';
    fft_set_int:='';
    help_int:=0;
    minbits_int:=-1;
    forcing_int:=0;
    fft32_int:=0;
    correct_int:=0;
    checking_int:=0;
    merging_int:=0;
    noclips_int:=0;
    analyses_int:=0;
    verbosity_int:=1;
    snr_int:=-216;
    lowpass_int:=-1;
    warnings_int:=1;
    shaping_int:=-1;
    autoshape_int:=-1;
    skewing_int:=-1;
    scaling_int:=-1;
    noisethreshold_int:=255;
    quality_int:=-1;
    priority_int:=0;
    detail_int:=0;
    dither_int:=0;
    silent_int:=0;
    WAVOutDir:='';

    wavName:='';
    initNextParamStr;

    if nextParamStr(s)=0 then
      exit;

    if (s[1]='-') then
    Begin
      if (s='-v') or (s='--version') then
        writeversionstring
      else
      begin
        write(lossyWAVhead1,crlf);
        write(lossyWAVhead3,crlf);
        if (s='-h') or (s='--help') or (s='-L') or (s='--longhelp') then
        begin
          help_int:=1+ord((s='--longhelp') or (s='-L'));
          parmserror;
          halt(0);
        end
        else
          lossyWAVerror('No input file given.',1);
      end;
    End;

    wavName:=s;
    if (wavName[1]='"') then
      wavName:=copy(wavName,2,length(wavName)-1);
    if (wavName[length(wavName)]='"') then
      wavName:=copy(wavName,1,length(wavName)-1);

    colonpos:=pos(':',wavName);
    if colonpos>0 then
    begin
      wavInpDir:=copy(wavName,1,colonpos);
      wavName:=copy(wavName,colonpos+1,length(wavName)-colonpos);
    end
    else
      wavInpDir:='';

    slashpos:=pos('\',wavName);
    while slashpos>0 do
    begin
      wavInpDir:=wavInpDir+copy(wavName,1,slashpos);
      wavName:=copy(wavName,slashpos+1,length(wavName)-slashpos);
      slashpos:=pos('\',wavName);
    end;

    while nextParamStr(s)=1 do
    begin
      if not check_quality(s) then
        if not check_system(s) then
          if not check_output(s) then
            if not check_file_output(s) then
              if not check_advanced(s) then
              begin
                parmError:='%lossyWAV Error% : Incorrect option: "' + s + '"'+crlf;
                exit;
              end;
    end;
    getParms:=1;
  end;

  function DirInp : String;
  begin
    DirInp:=wavInpDir;
  end;

  function DirOut : String;
  begin
    DirOut:=wavOutDir;
  end;

  Function wavFileName : String;
  Begin
    wavFileName:=wavName;
  End;

  Function lossyFileName : String;
  Begin
    lossyFileName:=lossyName;
  End;

  Function lwcdfFileName : String;
  Begin
    lwcdfFileName:=lwcdfName;
  End;

  function wavInp: string;
  begin
    wavInp:=wavInpDir + wavName;
  end;

  Function WAVOut: String;
  Begin
    WAVOut:=WAVOutDir + WAVName;
  End;

  function lossyInp: String;
  begin
    lossyInp:=WAVInpDir + lossyFileName;
  end;

  function lossyOut: string;
  begin
    lossyOut:=WAVOutDir + lossyFileName;
  end;

  function lwcdfInp: String;
  begin
    lwcdfInp:=WAVInpDir + lwcdfFileName;
  end;

  function lwcdfOut: string;
  begin
    lwcdfOut:=WAVOutDir + lwcdfFileName;
  end;

  function priority : Integer;
  begin
    priority:=priority_int;
  end;

  function noisethreshold: Double;
  begin
    noisethreshold:=noisethreshold_int;
  end;

  function quality: Double;
  begin
    quality:=quality_int;
  end;

  function forcing : Integer;
  begin
    forcing:=forcing_int;
  end;

  function fft32 : Integer;
  begin
    fft32:=fft32_int;
  end;

  function correct : Integer;
  begin
    correct:=correct_int;
  end;

  function checking : Integer;
  begin
    checking:=checking_int;
  end;

  function noclips : Integer;
  begin
    noclips:=noclips_int;
  end;

  function merging : Integer;
  begin
    merging:=merging_int;
  end;

  function dither : Integer;
  begin
    dither:=dither_int;
  end;

  function silent : Integer;
  begin
    silent:=silent_int;
  end;

  function analyses : Integer;
  begin
    analyses:=analyses_int;
  end;

  function minbits : Double;
  begin
    minbits:=minbits_int;
  end;

  function snr : Double;
  begin
    snr:=snr_int;
  end;

  function verbosity : Integer;
  begin
    verbosity:=verbosity_int;
  end;

  function detailledoutput : Integer;
  begin
    detailledoutput:=detail_int;
  end;

  function warnings : Integer;
  begin
    warnings:=warnings_int;
  end;

  function shaping : Double;
  begin
    shaping:=shaping_int;
  end;

  function autoshape : Integer;
  begin
    autoshape:=autoshape_int;
  end;

  function skewing : Double;
  begin
    skewing:=skewing_int;
  end;

  function lowpass : Double;
  begin
    lowpass:=lowpass_int;
  end;

  function scaling : Double;
  begin
    scaling:=scaling_int;
  end;

  function spf_string : String;
  begin
    spf_string:=spf_int;
  end;

  function fft_string : String;
  begin
    fft_string:=fft_set_int;
  end;

//==============================================================================
begin
//==============================================================================
  if (getparms=0) then
  begin
    write(lossyWAVhead1,crlf);
    write(lossyWAVhead3,crlf);
    if parmerror<>'' then
      write(parmerror);
    write('Syntax: lossyWAV <wavfilename>',crlf);
    write(crlf,'Type ''lossyWAV --help'' for more detail.');
    halt(1);
  end;

  if (help_int>=1) then
  begin
    write(lossyWAVhead1,crlf);
    write(lossyWAVhead3,crlf);
    parmserror;
    halt(1);
  end;

  if (verbosity=1) and (silent=0) then
    write(lossyWAVheader);

  if (length(wavName)=0) then
    lossyWAVerror('Name of input or output wav file missing.',$31);

  if not fileexists(wavInp) then
  begin
    if not fileexists(wavInp+'.wav') then
      lossyWAVerror('Input file: '+wavName+' does not exist.',$31)
    else
      wavName:=wavName+'.wav';
  end;
  lossyName:=copy(wavName, 1, length(wavName) - 3) + 'lossy.' + copy(wavName, length(wavName) - 2, 3);
  lwcdfName:=copy(wavName, 1, length(wavName) - 3) + 'lwcdf.' + copy(wavName, length(wavName) - 2, 3);

  if (WAVOutdir<>'') and (directoryexists(WAVOutdir)=false) then
    lossyWAVerror('Directory : '+WAVOutdir+' cannot be accessed.',$31);

  if merging=0 then
  begin
    if (fileexists(lossyOut)) then
    begin
      if forcing=1 then
      Begin
        if filegetattr(lossyOut) and 1=1 then
        Begin
          lossyWAVwarning('Output lossy file is read-only.');
          if(filesetattr(lossyOut,filegetattr(lossyOut) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+lossyOut,$31);
        End;
      End
      else
        lossyWAVerror('Output file: '+lossyOut+' exists.',$31);
    end;

    if (correct=1) and (fileexists(lwcdfOut)) then
    begin
      if forcing=1 then
      Begin
        if filegetattr(lwcdfOut) and 1=1 then
        Begin
          lossyWAVwarning('Output correction file is read-only.');
          if(filesetattr(lwcdfOut,filegetattr(lwcdfOut) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+lwcdfOut,$31);
        End;
      End
      else
        lossyWAVerror('Output file: '+lwcdfOut+' exists.',$31);
    end;
  end
  else
  begin
    lossyName:=wavName;
    if (copy(lossyName,length(lossyName)-9,10)='.lossy.wav') then
    begin
      lwcdfName:=copy(wavName,1,length(wavName)-10)+'.lwcdf.wav';
      wavName:=copy(wavName,1,length(wavName)-10)+'.wav';
    end;

    if fileexists(WAVOut) and (forcing=0) then
      lossyWAVerror('Output merged file already exists',$31);

    if not fileexists(lossyInp) then
      lossyWAVerror('Input lossy file not found.',$31);

    if not fileexists(lwcdfInp) then
      lossyWAVerror('Input correction file not found.',$31);
  end;

//==============================================================================
// Unit nparameter end;
//==============================================================================
  nparameter_dummy_integer[0]:=0;
end.
//==============================================================================

