//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nwav;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8 $I-}
//==============================================================================
interface
//==============================================================================
uses
  ntypes;

const
  maxchannels      = 8;
  maxblocksize     = 4096;

  CUEChunkMaxSize  = 65536;
  FACTChunkMaxSize = 65536;
  FMTChunkMaxSize  = 18+65536;

  bitstocalculate  = 32;
  bytestocalculate = 4;
  buffsize         = maxblocksize*maxchannels*bytestocalculate*2;

  channel_offset : array[0..maxchannels-1] of integer = (0,integer_size*1*maxblocksize,integer_size*2*maxblocksize,integer_size*3*maxblocksize,integer_size*4*maxblocksize,integer_size*5*maxblocksize,integer_size*6*maxblocksize,integer_size*7*maxblocksize);

Type
  allChannelsSampleBlock = array[0..maxchannels-1, 0..maxblocksize-1] of Integer;
  allChannelsSampleBlockPtr = ^AllChannelsSampleBlock;

  tChunkID = array[0..3] of Char;

  tChunkHeader = packed record
    Case integer of
     0 : (ID:tChunkID; Size : Cardinal);
     1 : (CNum : Integer);
    End;

  tChunkMapRecord = Packed Record
    Header : tChunkHeader;
    CPtr   : Integer;
  End;

  tWAVHeader = Packed Record
    Header   : tChunkHeader; //"RIFF"
    RIFFType : tChunkID; //"WAVE"
  end;

  tFMTChunk = Packed Record
    Header         : tChunkHeader; //"fmt "
    Compression    : Word;
    Channels       : Word;
    SampleRate     : Cardinal;
    AverageBPS     : Cardinal;
    BlockAlign     : Word;
    BitsPerSample  : Word;
    ExtraBytesSize : Word;
    ExtraBytes     : array[0..65535] of Byte;
  End;

  tFACTChunk = Packed Record
    Header : tChunkHeader; //"fact"
    Data : array[0..65535] of char;
  End;

  tDATAChunk = Packed Record
    Header : tChunkHeader; //"data"
  End;

  tlossyWAVChunk = Packed Record
   WAV  : tWAVHeader;
   FMT  : tFMTChunk;
   FACT : tFACTChunk;
   DATA : tDataChunk;
  End;

  tWaveBuffer = array[0..Buffsize-1] of byte;

var
  btrd_codec_block : AllChannelsSampleBlock;
  corr_codec_block : AllChannelsSampleBlock;

  rotating_blocks     : array[0..3] of AllChannelsSampleBlock;
  rotating_blocks_ptr : array[0..3] of AllChannelsSampleBlockPtr = (@rotating_blocks[0],@rotating_blocks[1],@rotating_blocks[2],@rotating_blocks[3]);

  prev_codec_block_size  : Integer = 0;
  last_codec_block_size  : Integer = 0;
  this_codec_block_size  : Integer = 0;
  next_codec_block_size  : Integer = 0;
  this_codec_block_size_recip : Double = 0;

function channelCount: integer;                             // number of channels
function sampleRate: Integer;                               // sample rate in Hz
function bitsPerSample: integer;                            // bits per sample
function bytesPerSample: integer;                           // bytes per sample
function totalsamples : integer;                            // total samples in WAV file.

function openWavIO(inWav: string; outWav: string; blocksize: Integer): integer;
                                                            // False if operation not succesful.
                                                            // Reads the header chunks of inWav and memorizes the
                                                            // Headerrmation from the fmt chunk.
                                                            // Creates outWav and fills its header chunks with an
                                                            // identical copy of that of inWav.
                                                            // Memorizes blocksize (nr of samples to read/write with subsequent calls to
                                                            // readNextSampleBlock resp. writeNextSampleBlock).

function readNextNextCodecBlock : integer;                  // False if not possible (usually because all the samples have been read).
                                                            // sampleBlock: next block of samples from input wav file for all the channels.
                                                            //              first sample is in element with index 0.
                                                            // readsamplesCount: number of samples read (in each of the channels).

function writeNextbtrdcodecblock : integer;                 // False if not possible.
                                                            // Writes sampleBlock as the next sample block of all the channels to the output file.
                                                            // The number of samples to be written is given by samplesCountToBeWritten.
                                                            // First sample written is sampleBlock element with index overlapsize
                                                            // - overlapsize from openWavIO call.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.

function writeNextcorrcodecblock : integer;                 // False if not possible.
                                                            // Writes sampleBlock as the next sample block of all the channels to the output file.
                                                            // The number of samples to be written is given by samplesCountToBeWritten.
                                                            // First sample written is sampleBlock element with index overlapsize
                                                            // - overlapsize from openWavIO call.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.

function closeWavIO : integer;                              // False if not possible.
                                                            // Finalizes the wavIO operations.

procedure MergeFiles;

//==============================================================================
implementation
//==============================================================================
uses
  nparameter,
  Math,
  SysUtils,
  Windows;

//==============================================================================
//nwav constants, variables & types
//==============================================================================
Var
  WAVBuff                             : tWaveBuffer;
  BtrdBuff                            : tWaveBuffer;
  Corrbuff                            : tWaveBuffer;
  MergeBuff                           : tWaveBuffer;
  chunkspace                          : array[0..2097151] of byte;

  WAVEFile, Btrdfile, Corrfile        : File;
  filesOpened                         : integer;

  readresult                          : Cardinal;
  writeresult                         : Cardinal;
  ChunkHeader                         : tChunkHeader;
  WAVEChunk                           : tlossyWAVChunk;
  BTRDChunk                           : tlossyWAVChunk;
  CORRChunk                           : tlossyWAVChunk;

  buffsizeread                        : Integer; // number of byte that make up for inbuff
  buffsizewrite                       : Integer; // number of byte that make up for outbuff
  nrOfByteInOneBlockInBuff            : Integer; // number of byte that make up a complete block in inBuff and outBuff.
  nrOfBlocksInBuff                    : integer; // inBuff size as a factor of blocksize
  nrOfFullBlocksReadIntoInBuff        : integer; // with the last BlockRead: how many full blocks managed it into inBuff ?
  nrOfByteOfTheLastIncompleteBlock    : Integer; // when EOF was encountered with the last BlockRead: the number of byte that didn't make it
                                                 // into the last full block
  nrOfBlockInBuffNotYetFetched        : integer; // the next number of the block in inBuff that was not fetched yet by readNextSampleBlock.
                                                 // 0: no data in inBuff;
  nrOfByteFilledInBtrdBuff            : Integer; // how many byte have managed it into outBuff that are not written to the outputfile yet?
  nrOfByteFilledInCorrBuff            : Integer; // how many byte have managed it into outBuff that are not written to the outputfile yet?
  bitdepth                            : integer;
  bytedepth                           : integer;
  sampleByteLeftToRead                : Integer;
  readbyteCount                       : Integer;
  CantReadInWavAnyMore                : integer;

  datetimestring                      : string;

  chunkmap                            : array[0..1023] of tChunkMapRecord;
  numchunks                           : integer = 0;
  endofchunkmap : Integer;

  last_find_text                      : integer = -1;
  this_scale                          : Double;
  this_scale_inv                      : Double;
  buffsizetouse                       : Cardinal;
  totalsamples_int                    : Integer;

  nwav_packer                         : array[0..2] of integer;

  function channelCount: integer;
  begin
    channelCount:=WAVEChunk.FMT.Channels;
  end;

  function sampleRate: Integer;
  begin
    sampleRate:=WAVEChunk.FMT.SampleRate;
  end;

  function bitsPerSample: integer;
  begin
    bitsPerSample:=bitdepth;
  end;

  function bytesPerSample: integer;
  begin
    bytesPerSample:=bytedepth;
  end;

  function totalsamples : integer;
  begin
    totalsamples:=totalsamples_int;
  end;

  procedure combinelossybuffers(bytes_in_buffer : Integer);
  var
    this_sample, btrd_sample, corr_sample : integer;
  begin
    case bytedepth of
      1 : lossyWAVerror('Invalid bitdepth.',$12);

      2 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,2
                cmp    ecx,0
                jl     @@01

                movsx  eax,word ptr [ebx+ecx]
                mov    dword ptr btrd_sample,eax

                movsx  eax,word ptr [edi+ecx]
                mov    dword ptr corr_sample,eax

                fild   btrd_sample
                fmul   this_scale_inv
                fiadd  corr_sample
                fistp  this_sample

      @@00b:    mov    eax,this_sample
                mov    word ptr [esi+ecx],ax
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
          end;

      3 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,3
                cmp    ecx,0
                jl     @@01

                xor    eax,eax
                movsx  ax,byte ptr [ebx+ecx+2]
                shl    eax,16
                mov    ax,word ptr [ebx+ecx]
                mov    dword ptr btrd_sample,eax

                xor    eax,eax
                movsx  ax,byte ptr [edi+ecx+2]
                shl    eax,16
                mov    ax,word ptr [edi+ecx]
                mov    dword ptr corr_sample,eax

                fild   btrd_sample
                fmul   this_scale_inv
                fiadd  corr_sample
                fistp  this_sample

                mov    eax,this_sample
                mov    word ptr [esi+ecx],ax
                shr    eax,16
                mov    byte ptr [esi+ecx+2],al
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
          end;

      4 : asm
                PushF
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,4
                cmp    ecx,0
                jl     @@01

                fild   dword ptr [ebx+ecx]
                fmul   this_scale_inv
                fiadd  dword ptr [edi+ecx]
                fistp  dword ptr [esi+ecx]
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
                popf
         end;
    end;
  end;

  procedure openfile(var thisfile : file; thisname : string; thisfilemode : Integer; orvalue : integer);
  begin
    filemode:=thisfilemode;

    AssignFile(thisfile, thisname);
    if IOResult <> 0 then exit;

    case thisfilemode of
      0 : Reset(thisfile, 1);
      2 : Rewrite(thisfile,1);
    else
      lossyWAVerror('Invalid FileMode!',$12);
    end;
    if IOResult <> 0 then exit;

    filesOpened:=filesopened or orvalue;
  end;

  procedure wavIOExitProc(wavIOString : String; wavIOCode : Integer);
  var
    wi : integer;
  begin
    for wi:=0 to numchunks-1 do
      write(chunkmap[wi].Header.ID,',');
    write(crlf);
    write(wavIOString,crlf);
    closeWavIO;
    exitcode:=wavIOcode;
    halt;
  end;

  procedure writeWAVheaders(var outputfile : file);
  var
    thischunk : integer;
  begin
    if merging=1 then
      WAVEChunk.WAV.Header.Size:=WAVEChunk.WAV.Header.Size-(WAVEChunk.FACT.Header.Size+8);

    BlockWrite(outputfile, WAVEChunk.WAV, 12);
    if IOResult <> 0 then exit;

    for thischunk:=0 to numchunks-1 do
    begin
      if chunkmap[thischunk].cptr=-1 then
      begin
        case chunkmap[thischunk].header.cnum of
          $20746D66 : begin //'fmt '
                        BlockWrite(outputfile, WAVEChunk.FMT, WAVEChunk.FMT.Header.Size+8);
                        if IOResult <> 0 then
                          wavIOExitProc('Error writing ''fmt '' chunk.',$12);
                        if (merging=0) and (WaveChunk.Fact.Header.Size<>0) then
                        begin
                          BlockWrite(outputfile, WAVEChunk.FACT, WAVEChunk.FACT.Header.Size+8);
                          if IOResult <> 0 then
                            wavIOExitProc('Error writing ''fact'' chunk.',$12);
                        end;
                      end;
          $74636166 : begin
                      end;
        end
      End
      else
      Begin
        Blockwrite(outputfile,chunkmap[thischunk].header,8);
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
        Blockwrite(outputfile,chunkspace[chunkmap[thischunk].cptr],chunkmap[thischunk].header.size);
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
      end;
    end;

    BlockWrite(outputfile, WAVEChunk.DATA, 8);
    if IOResult <> 0 then exit;
  end;

  function HexChar(HC : Integer) : Char;
  Asm
                Mov    eax,hc
                and    eax,15
                cmp    eax,9
                jle    @@01
                add    eax,7
      @@01:     add    eax,48
  End;

  function Dec2Hex(d2h : integer) : String;
  Begin
    Dec2Hex:=HexChar(d2h shr 28)+HexChar(d2h shr 24)+HexChar(d2h shr 20)+HexChar(d2h shr 16)+HexChar(d2h shr 12)+HexChar(d2h shr 8)+HexChar(d2h shr 4)+HexChar(d2h)
  End;

  procedure OpenAndReadChunks(ThisName : String; var ThisFile : File; var ThisChunk : tlossyWAVChunk; thisorvalue : Integer);
  begin
    openfile(ThisFile,ThisName,0,thisorvalue);
    fillchar(thischunk,sizeof(thischunk),0);
    numchunks:=0;
    endofchunkmap:=0;

    //==========================================================================

    BlockRead(ThisFile, ThisChunk.WAV, 12, readbyteCount);
    if (IOResult <> 0) or (readbyteCount <> 12) then
      wavIOExitProc('Error reading WAV header.',$12);

    //==========================================================================

    if ThisChunk.WAV.Header.ID <> 'RIFF' then
      wavIOExitProc('File type is not RIFF.',$12);

    if ThisChunk.WAV.RIFFType <> 'WAVE' then
      wavIOExitProc('RIFF type is not WAVE.',$12);

    //==========================================================================

    repeat
      BlockRead(ThisFile,ChunkHeader,8,readresult);
      if readresult<>8 then
        wavIOExitProc('Error Reading WAV file Information.',$11);

      case chunkheader.cnum of
        $20746D66 : begin //'fmt '
                      chunkmap[numchunks].Header:=chunkheader;
                      chunkmap[numchunks].Cptr:=-1;
                      numchunks:=numchunks+1;
                      move(chunkheader,ThisChunk.FMT,8);
                      if ThisChunk.FMT.Header.Size>FMTChunkMaxSize then
                        WavIOExitProc('FMT Chunk Too Large.',$12);
                      blockread(ThisFile,ThisChunk.FMT.Compression,thischunk.FMT.Header.Size,readresult);
                      if readresult<>thischunk.FMT.Header.Size then
                        wavIOExitProc('Error reading FMT chunk.',$12);
                    end;
        $61746164 : begin //'data'
                      Move(ChunkHeader,ThisChunk.DATA,8);
                    end;
      else
        chunkmap[numchunks].Header:=chunkheader;
        chunkmap[numchunks].Cptr:=EndofChunkMap;
        If endofchunkmap>sizeof(chunkspace) then
          wavIOExitProc('Too much WAV information before ''data'' chunk!',$12);
        blockread(thisfile,chunkspace[endofchunkmap],chunkheader.size,readresult);
        if readresult<>chunkheader.size then
          wavIOExitProc('Error Reading '''+chunkheader.ID+''' Chunk.',$12);
        endofchunkmap:=endofchunkmap+integer(chunkheader.size);

        if (chunkmap[numchunks].Header.Id='fact') then
        begin
          ThisChunk.FACT.Header:=ChunkMap[numchunks].Header;
          if thischunk.fact.header.size<factchunkMaxsize then
            move(chunkspace[chunkmap[numchunks].Cptr],thischunk.fact.data,thischunk.fact.header.size);
          if copy(thischunk.fact.data,1,8)<>'lossyWAV' then
            thischunk.fact.header.size:=0
          else
            chunkmap[numchunks].cptr:=-1;
        end;

        numchunks:=numchunks+1;
      end;
    until ChunkHeader.ID='data';

    //==========================================================================

    if (ThisChunk.FMT.Compression<>1) then
      wavIOExitProc('WAV Data is not uncompressed PCM.',$12);
  end;

  function find_text(find_str : string; source_str : string) : integer;
  var
    ft_i : Integer;
    ft_l : Integer;
  begin
    ft_i:=0;
    ft_l:=length(find_str);

    if ft_l=0 then
      ft_i:=-1
    else
    begin
      repeat
        inc(ft_i);
      until (ft_i=(length(source_str)-ft_l)) or (find_str=copy(source_str,ft_i,ft_l));

      if ft_i=(length(source_str)-ft_l) then
        ft_i:=-1
    end;
    find_text:=ft_i;
    last_find_text:=ft_i;
  end;

  procedure MergeFiles;
  var
    thisread   : cardinal;
    numreads   : cardinal;
    lastread   : cardinal;
    readcount  : cardinal;
    this_error : Integer;
    temp_str   : String;
  begin
    filesOpened:=0;

    OpenAndReadChunks(LossyInp,BTRDFile,BTRDChunk,2);

    OpenAndReadChunks(lwcdfInp,CORRFile,CORRChunk,4);

    Write(crlf,'Merging: ',lossyfilename,crlf);
    Write(     '       & ',lwcdffilename,crlf);

    if btrdchunk.wav.Header.Size<>corrchunk.wav.Header.Size then
      lossyWAVerror('Files not same length.',$12);

    if btrdchunk.data.Header.Size<>corrchunk.data.Header.Size then
      lossyWAVerror('Data not same length.',$12);

    if btrdchunk.fact.data<>corrchunk.fact.data then
      lossyWAVerror('FACT Data mismatch.',$12);

    if copy(btrdchunk.fact.data,1,8)<>'lossyWAV' then
      lossyWAVerror('Not a lossyWAV FACT chunk.',$12);

    if (btrdchunk.fmt.Header.Size<>corrchunk.fmt.Header.Size) or
       (btrdchunk.fmt.Compression<>corrchunk.fmt.compression) or
       (btrdchunk.fmt.Channels<>corrchunk.fmt.Channels) or
       (btrdchunk.fmt.SampleRate<>corrchunk.fmt.SampleRate) or
       (btrdchunk.fmt.AverageBPS<>corrchunk.fmt.AverageBPS) or
       (btrdchunk.fmt.BlockAlign<>corrchunk.fmt.BlockAlign) or
       (btrdchunk.fmt.BitsPerSample<>corrchunk.fmt.BitsPerSample) then
      lossyWAVerror('FMT Data mismatch.',$12);

    bytedepth:=(btrdchunk.fmt.bitspersample+7) shr 3;
    if bytedepth<2 then
      lossyWAVerror('Invalid bitdepth: '+inttostr(btrdchunk.fmt.bitspersample),$12);

    lastread:=btrdChunk.Data.Header.Size mod buffsize;
    numreads:=btrdchunk.data.Header.Size div buffsize+cardinal(lastread>0);

    move(btrdchunk,wavechunk,sizeof(btrdchunk));
    wavechunk.wav.Header.Size:=wavechunk.wav.Header.Size-(wavechunk.fact.Header.Size+8);
    openfile(WAVEFile,WAVOut,2,1);

    blockwrite(WAVEFile,wavechunk.wav,12,writeresult);
    if writeresult<>12 then
      lossyWAVerror('Writing WAV Header Chunk.',$12);
    blockwrite(WAVEFile,wavechunk.fmt,wavechunk.fmt.Header.Size+8,writeresult);
    if writeresult<>wavechunk.fmt.Header.Size+8 then
      lossyWAVerror('Writing WAV FMT Chunk.',$12);
    blockwrite(WAVEFile,wavechunk.data,8,writeresult);
    if writeresult<>8 then
      lossyWAVerror('Writing WAV DATA Header.',$12);

    if find_text('-scale ',wavechunk.fact.data)<>-1 then
      temp_str:=copy(wavechunk.fact.data,last_find_text+7,8)
    else
      temp_str:='1.000000';

    val(temp_str,this_scale,this_error);
    if this_error<>0 then
      lossyWAVerror('Invalid scale value: '+temp_str,$12);

    this_scale_inv:=1/this_scale;

    buffsizetouse:=buffsize-(buffsize mod (bytedepth*btrdchunk.fmt.channels));

    for readcount:=0 to numreads-1 do
    begin
      thisread:=min(buffsizetouse,btrdchunk.data.Header.Size-Readcount*buffsizetouse);
      blockread(BTRDFile,BTRDBuff,thisread,readresult);
      blockread(CORRFile,CORRBuff,thisread,readresult);
      readbytecount:=readbytecount+integer(readresult);
      combinelossybuffers(thisread);
      blockwrite(WAVEFile,MERGEBuff,thisread,readresult);
      write(#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    close(CORRFile);
    readbytecount:=filepos(btrdfile);
    readresult:=65536;

    while readresult=65536 do
    begin
      blockread(BTRDFile,BTRDBuff,65536,readresult);
      blockwrite(WAVEFile,btrdbuff,readresult);
      readbytecount:=readbytecount+integer(readresult);
      write(#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    write(crlf);

    close(BTRDFile);
    close(WAVEFile);
  end;

  function openWavIO(inWav: string; outWav: string; blocksize: Integer): integer; // only standard chunks assumed so far
  begin
    openWavIO:=0;
    filesOpened:=0;

    OpenAndReadChunks(inwav,WAVEFile,WAVEChunk,1);

    bitdepth:=WAVEChunk.FMT.BitsPerSample;
    if (bitdepth<9) or (bitdepth>32) then
    begin
      write('Input file has a bitdepth of ' + IntToStr(bitdepth) + ' bits per sample.'+crlf);
      wavIOExitProc('This is not supported.',$12);
    end;
    bytedepth:=(bitdepth+7) shr 3;

    totalsamples_int:=floor(WAVEChunk.DATA.Header.Size / WAVEChunk.FMT.Channels / bytedepth);

    if samplerate<32000 then
    begin
      write('Sample Rate too low : '+inttostr(samplerate)+'Hz (Min=32000Hz).',crlf);
      wavIOExitProc('This is not supported.',$12);
    end;

    sampleByteLeftToRead:=WAVEChunk.DATA.Header.Size;
    nrOfByteInOneBlockInBuff:=bytedepth * WAVEChunk.FMT.Channels * blocksize;
    nrOfBlocksInBuff:=buffsize div nrOfByteInOneBlockInBuff;
    buffsizeread:=nrOfByteInOneBlockInBuff * nrOfBlocksInBuff;
    buffsizewrite:=buffsize;
    CantReadInWavAnyMore:=0;

    //==========================================================================

    if checking=1 then
    begin
      if (WAVEChunk.FACT.Header.Size>0) and (copy(WAVEChunk.FACT.Data,1,8)='lossyWAV') then
      begin
        exitcode:=$10;
        if (verbosity=1) and (silent=0) then
        begin
          write('lossyWAV FACT Chunk found. File already processed.',crlf);
          write(WAVEChunk.FACT.Data);
        end;
      end
      else
      begin
        exitcode:=$00;
        if (verbosity=1) and (silent=0) then
          write('lossyWAV FACT Chunk not found. File not marked as processed.',crlf);
      end;
      closeWAVio;
      halt;
    end
    else
    begin
      if copy(WAVEChunk.FACT.data,1,8)='lossyWAV' then
      begin
        write(WAVEChunk.FACT.Data);
        lossyWAVerror('lossyWAV FACT Chunk found. File already processed.',$10);
      end;
    end;

    //==========================================================================
    //create 'fact' chunk data string.
    datetimetostring(datetimestring,'',now);
    WAVEChunk.FACT.Header.ID:='fact';
    DateTimeString:=version_string+' @ '+datetimestring+', '+parameter_string;
    if length(DateTimestring) and 1=1 then
      DateTimestring:=DateTimestring+#0;
    move(DateTimestring[1],WAVEChunk.FACT.data,length(DateTimestring));
    WAVEChunk.FACT.Header.Size:=length(DateTimestring);
    inc(WAVEChunk.WAV.Header.Size,WAVEChunk.FACT.Header.Size+8);

    BlockRead(WAVEFile, WAVBuff, min(samplebytelefttoread,buffsizeread), readbyteCount); // pre-read into inbuff
    if IOResult <> 0 then
    begin
      CantReadInWavAnyMore:=1;
      exit;
    end;

    if readbyteCount = 0 then
    begin
      CantReadInWavAnyMore:=1;
      exit; // no data
    end;

    sampleByteLeftToRead:=sampleByteLeftToRead - readByteCount;
    nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
    nrOfByteOfTheLastIncompleteBlock:=readbyteCount - nrOfFullBlocksReadIntoInBuff * nrOfByteInOneBlockInBuff;
    nrOfBlockInBuffNotYetFetched:=1;
    nrOfByteFilledInBtrdBuff:=0;
    nrOfByteFilledInCorrBuff:=0;

    //==========================================================================

    openfile(BtrdFile,lossyout,2,2);
    writeWAVheaders(btrdfile);

    if correct=1 then
    begin
      openfile(CorrFile,lwcdfout,2,4);
      writeWAVheaders(Corrfile);
    end;

    //==========================================================================

    openWavIO:=1;
  end;

  procedure transfer(sampleBlockTransferStart: Integer; pStartB: PByte; pEndB: PByte);
  begin
    case bytedepth of
      2 : asm
                push   ebx
                push   edi

                mov    edi,sampleblocktransferstart
                mov    ecx,pStartB

      @@20:     mov    ebx,0

      @@21:     movsx  eax,word ptr [ecx]
                add    ecx,2

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.Channels
                jl     @@21

                add    edi,1
                cmp    ecx,pEndB
                jle    @@20

                pop    edi
                pop    ebx
          end;

      3 : asm
                push   ebx
                push   edi

                mov    edi,sampleblocktransferstart
                mov    ecx,pStartB

      @@30:     mov    ebx,0

      @@31:     movsx  eax,byte ptr [ecx+2]
                shl    eax,16
                mov    ax,word ptr [ecx]
                add    ecx,3

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.Channels
                jl     @@31

                add    edi,1
                cmp    ecx,pEndB
                jle    @@30

                pop    edi
                pop    ebx
          end;

      4 : asm
                push   ebx
                push   edi

                mov    edi,sampleblocktransferstart
                mov    ecx,pStartB

      @@40:     mov    ebx,0

      @@41:     mov    eax,[ecx]
                add    ecx,4

                mov    edx,dword ptr channel_offset[ebx*pointer_size]
                add    edx,dword ptr rotating_blocks_ptr[3*pointer_size]
                mov    dword ptr [edx+edi*integer_size],eax

                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.Channels
                jl     @@41

                add    edi,1
                cmp    ecx,pEndB
                jle    @@40

                pop    edi
                pop    ebx
          end;
    end;
  end;

  function readNextNextCodecBlock : integer;
  var
    sampleBlockStart: Integer;
    pB, pStartB: PByte;
    lastNrOfBlockInBuffNotYetFetched: integer;
  begin
    readNextnextCodecBlock:=0;
    sampleBlockStart:=0;
    if nrOfBlockInBuffNotYetFetched = 0 then // samples to be fetched aren't in inbuff
    begin
      if nrOfFullBlocksReadIntoInBuff < nrOfBlocksInBuff then // EOF encountered with last call to readNextSampleBlock
        exit; // no more data
      if sampleByteLeftToRead = 0 then
        exit; // no more data
      sampleBlockStart:=0;

      BlockRead(WAVEFile, WAVBuff, min(buffsizeread,samplebytelefttoread), readbyteCount);
      if IOResult <> 0 then
      begin
        CantReadInWavAnyMore:=1;
        exit;
      end;

      if readbyteCount = 0 then
      begin
        CantReadInWavAnyMore:=1;
        exit; // no more data
      end;

      if readbyteCount < buffsizeread then
        CantReadInWavAnyMore:=1;

      sampleByteLeftToRead:=sampleByteLeftToRead - readByteCount;
      nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
      nrOfByteOfTheLastIncompleteBlock:=readbyteCount - nrOfFullBlocksReadIntoInBuff * nrOfByteInOneBlockInBuff;
      nrOfBlockInBuffNotYetFetched:=1;
    end;

    pB:=@WAVBuff;
    Inc(pB, (nrOfBlockInBuffNotYetFetched - 1) * nrOfByteInOneBlockInBuff);
    pStartB:=pB; // points to first sample in Block
    lastNrOfBlockInBuffNotYetFetched:=nrOfBlockInBuffNotYetFetched;

    if nrOfBlockInBuffNotYetFetched <= nrOfFullBlocksReadIntoInBuff then
    begin
      Inc(pB, nrOfByteInOneBlockInBuff);  // pB points to the spot after last sample in Block
      next_codec_block_size:=nrOfByteInOneBlockInBuff DIV (bytedepth * WAVEChunk.FMT.Channels);
      nrOfBlockInBuffNotYetFetched:=nrOfBlockInBuffNotYetFetched + 1;
      if nrOfBlockInBuffNotYetFetched > nrOfBlocksInBuff then
        nrOfBlockInBuffNotYetFetched:=0;
    end
    else
    begin
      if nrOfByteOfTheLastIncompleteBlock = 0 then
        exit; // no more data
      Inc(pB, nrOfByteOfTheLastIncompleteBlock);  // pB points to the spot after last sample in Block
      next_codec_block_size:=nrOfByteOfTheLastIncompleteBlock DIV (bytedepth * WAVEChunk.FMT.Channels);
      nrOfBlockInBuffNotYetFetched:=0; // force a try for a new Blockread with next call which will lead to a return False
    end;

    if lastNrOfBlockInBuffNotYetFetched > 1 then
      sampleBlockStart:=0;
    transfer(sampleBlockStart, pStartB, pB);
    readNextNextCodecBlock:=1;
  end;

  function writeNextCodecBlock(var outputfile : file; var OutBuff; var OutputCodecBlock : AllChannelsSampleBlock; var FilledOutputBuffer : integer): integer;
  var
    nrOfByteForOutBuff : Integer;
    pB: PByte;
  begin
    writeNextCodecBlock:=0;
    nrOfByteForOutBuff:=bytedepth * WAVEChunk.FMT.Channels * this_codec_block_size;

    if FilledOutputBuffer + nrOfByteForOutBuff > buffsizewrite then // not sufficient space in outBuff : flush outBuff
    begin
      BlockWrite(outputfile, OutBuff, FilledOutputBuffer);
      if IOResult <> 0 then exit;

      FilledOutputBuffer:=0;
    end;

    pB:=@OutBuff;
    Inc(pB, FilledOutputBuffer); // points to first free spot in outBuff

    case bytedepth of
      2 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    word ptr [ecx],ax
                add    ecx,2

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.Channels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;

      3 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    word ptr [ecx],ax
                shr    eax,16
                mov    byte ptr [ecx+2],al
                add    ecx,3

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.Channels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;

      4 : asm
                push   ebx
                push   edi
                mov    ecx,pb
                mov    edi,0

      @@00:     mov    ebx,0

      @@01:     mov    edx,dword ptr channel_offset[ebx*integer_size]
                mov    eax,outputcodecblock
                add    eax,edx
                mov    eax,dword ptr [eax+edi*integer_size]

                mov    [ecx],eax
                add    ecx,4

                add    ebx,1
                cmp    bx,WAVEChunk.FMT.Channels
                jl     @@01

                add    edi,1
                cmp    edi,this_codec_block_size
                jl     @@00

                mov    pb,ecx
                pop    edi
                pop    ebx
          end;
    end;
    FilledOutputBuffer:=FilledOutputBuffer + nrOfByteForOutBuff;
    writeNextCodecBlock:=1;
  end;

  function writeNextBtrdCodecBlock: integer;
  begin
    WriteNextBtrdCodecBlock:=WriteNextCodecBlock(BtrdFile,BtrdBuff,btrd_codec_block,nrofByteFilledInBtrdBuff);
  end;

  function writeNextCorrCodecBlock : integer;
  begin
    WriteNextCorrCodecBlock:=WriteNextCodecBlock(CorrFile,Corrbuff,corr_codec_block,nrofByteFilledInCorrBuff);
  end;

  function closeWavIO: integer;
  begin
    closeWavIO:=0;
    if filesOpened>0 then
    begin
      if (filesopened and 2=2) and (nrOfByteFilledInBtrdBuff > 0) then
      begin
        BlockWrite(BtrdFile, BtrdBuff, nrOfByteFilledInBtrdBuff);
        if IOResult <> 0 then exit;
      end;

      if (filesopened and 4=4) and (nrOfByteFilledInCorrBuff > 0) then
      begin
        BlockWrite(CorrFile, CorrBuff, nrOfByteFilledInCorrBuff);
        if IOResult <> 0 then exit;
      end;

      CantReadInWavAnyMore:=0;
      While CantReadInWAVAnyMore=0 do
      Begin
        BlockRead(WAVEFile, WAVBuff, buffsizeread, readbyteCount);
        if IOResult <> 0 then
        begin
          CantReadInWavAnyMore:=1;
          break; // no more data
        end;

        if readbyteCount = 0 then
        begin
          CantReadInWavAnyMore:=1;
          break; // no more data
        end;

        if readbyteCount < buffsizeread then
          CantReadInWavAnyMore:=1;

        if filesopened and 2=2 then
        begin
          BlockWrite(BtrdFile, WAVBuff, readbyteCount);
          if IOResult <> 0 then exit;
        end;

        if filesopened and 4=4 then
        Begin
          BlockWrite(CorrFile, WAVBuff, readbyteCount);
          if IOResult <> 0 then exit;
        End;
      End;

      closeFile(WAVEFile);
      if IOResult <> 0 then exit;

      if filesopened and 2=2 then
      begin
        CloseFile(BtrdFile);
        if IOResult <> 0 then exit;
      end;

      if filesopened and 4=4 then
      begin
        closeFile(CorrFile);
        if ioresult<>0 then exit;
      end;
    end;

    closeWavIO:=1;
  end;

//==============================================================================
// Unit nwav end.
//==============================================================================
begin
  nwav_packer[0]:=0;
end.
//==============================================================================

